# encoding: UTF-8
require 'openssl'
require 'weakref'
require 'digest/sha1'

require_relative 'pem_scanner'

module SSLTool
  class Certificate < OpenSSL::X509::Certificate
    RX_DOMAIN_NAME = /^(\*\.)?([a-zA-Z0-9-]+\.)+[a-zA-Z0-9]+$/

    # The certificate_cache stuff ensures we always get the same object every time we instantiate the same certificate.
    # This is regardless of differences in the string used to instantiate the object.
    # The weakrefs ensure we don't hold on to certs that are not being used outside of the cache.
    @@certificate_cache = {}
    def self.new(s)
      cert = super(s)
      k = cert.fingerprint
      @@certificate_cache.delete(k) if v = @@certificate_cache[k] and v.respond_to?(:weakref_alive?) && !v.weakref_alive?
      (@@certificate_cache[k] ||= WeakRef.new(cert)).__getobj__
    end

    # returns an array of Certificate objects created from cert strings found in s
    def self.scan(s)
      PEMScanner.certs_from(s).uniq
    end

    ### signing

    def signed_by?(other_cert)
      verify(other_cert.public_key)
    rescue OpenSSL::X509::CertificateError => e
      OpenSSL.errors  # clear error queue to prevent pg errors
      return false if e.message == "wrong public key type"                                              # self.signature_algorithm is incompatible with type of other_cert.public_key; verify is definitely false
      return nil   if e.message == "unknown message digest algorithm" && signature_algorithm =~ /^md2/  # md2 is not present in later versions of openssl; can't tell signature verifies, so returning nil
      raise e
    end

    def signs?(other_cert)
      other_cert.signed_by?(self)
    end

    def self_signed?
      signs?(self)
    end

    def self_issued?
      subject.eql?(issuer)
    end

    def parent?(other_cert)
      signs?(other_cert) \
      && subject == other_cert.issuer \
      && subject_key_identifier == other_cert.authority_key_identifier
    end

    ### properties

    def fingerprint
      @fingerprint ||= Digest::SHA1.hexdigest(to_der)
    end

    def common_names
      subject.to_a.select { |k, _, _| k == "CN" }.map { |_, v, _| v }
    end

    def domain_names_from_common_names
      common_names.select { |cn| cn =~ RX_DOMAIN_NAME }
    end

    def domain_names_from_subject_alt_names
      map_extension_value('subjectAltName') { |s| s.scan(/\bDNS:([^\s,]+)/) }
    end

    def domain_names
      [ domain_names_from_common_names,
        domain_names_from_subject_alt_names,
      ].flatten.compact.sort.uniq
    end

    def for_domain_name?
      !domain_names.empty?
    end

    def certificate_authority?
      map_extension_value('basicConstraints') { |s| s.split(", ").include?('CA:TRUE') }
    end

    def certificate_sign?
      map_extension_value('keyUsage') { |s| s.split(", ").include?('Certificate Sign') }
    end

    def subject_key_identifier
      map_extension_value('subjectKeyIdentifier') { |s| s.chomp }
    end

    def authority_key_identifier
      map_extension_value('authorityKeyIdentifier') { |s| s.sub(/^keyid:/, '').chomp }
    end

    def key_size
      case public_key
      when OpenSSL::PKey::RSA ; public_key.n.num_bits
      when OpenSSL::PKey::EC  ; public_key.group.order.num_bits
      else                    ; nil
      end
    end

    def key_size_secure?
      case public_key
      when OpenSSL::PKey::RSA ; key_size >= 2048
      when OpenSSL::PKey::EC  ; key_size >= 224
      else                    ; nil
      end
    end

    def valid?
      now = Time.now.utc
      now <= not_after && now >= not_before
    end

    def acceptable?
      valid? && (!key_size || key_size_secure?)
    end

    ### extensions

    module Extensions
      def [](k)
        return super if k.is_a?(Integer) || !k.respond_to?(:to_str)
        find { |e| e.oid == k.to_str }
      end
    end

    def extensions
      super.tap { |a| class << a; include Extensions; end }
    end

    def map_extension_value(extension_name, default = nil)
      e = extensions[extension_name]
      return default if e.nil?
      block_given? ? yield(e.value) : e.value
    end

    ### chain

    def chain_from(certs)
      chain  = [self]
      parent = (certs - chain).find { |cert| cert.parent?(self) }
      chain.concat(parent.chain_from(certs - chain)) if parent
      chain
    end

  end
end
