# encoding: UTF-8

require_relative "error"
require_relative "certificate"
require_relative "certificate_store"

module SSLTool
  class ChainResolution
    class ChainResolutionError             < ::SSLTool::Error;     end
    class ZeroCertsChainResolutionError    < ChainResolutionError; end
    class ZeroHeadsChainResolutionError    < ChainResolutionError; end
    class TooManyHeadsChainResolutionError < ChainResolutionError
      attr_reader :domain_certs
      def initialize(domain_certs)
        @domain_certs = domain_certs
        super("#{@domain_certs.length} certificates given cover domain names and could be used as the chain head.")
      end
    end

    class CertificateBundle < Array
      alias_method :to_pem, :join
    end

    def is_trusted_root? cert
      cert.certificate_authority? &&
      cert.self_signed?           &&
      @certificate_store.trusted_pool.include?(cert)
    end

    def initialize(original_chain, certificate_store)
      @original_chain             = CertificateBundle.new(original_chain.uniq).freeze
      @certificate_store          = certificate_store
      @domain_certs, @other_certs = @original_chain.partition(&:for_domain_name?)
      @original_chain.empty? and raise ZeroCertsChainResolutionError
      case @domain_certs.length
      when 1; # pass
      when 0; raise ZeroHeadsChainResolutionError
      else  ; raise TooManyHeadsChainResolutionError.new(@domain_certs)
      end
      @base_cert             = @domain_certs.first
      @ordered_chain         = CertificateBundle.new(@base_cert.chain_from(@other_certs)).freeze
      @resolved_chain        = CertificateBundle.new(@base_cert.chain_from(@certificate_store.combined_trusted_pool_set))
                                 .take_while { |c| ! is_trusted_root? c }
                                 .freeze
      @unused_certs          = CertificateBundle.new(@other_certs - @resolved_chain).freeze
      @domain_names          = @base_cert.domain_names.freeze
      @originally_ordered    = @original_chain == @ordered_chain
      @originally_trusted    = @certificate_store.trust? @ordered_chain
      @ordered               = true
      @trusted               = @certificate_store.trust? @resolved_chain
      @self_signed_untrusted = @resolved_chain.last.self_signed? && !@trusted
      @recommended_chain     = case \
        when @originally_ordered && @originally_trusted ; @original_chain
        when @originally_trusted                        ; @ordered_chain
        else                                            ; self
        end
    end

    attrs = %w[ base_cert domain_names original_chain ordered_chain resolved_chain recommended_chain unused_certs ]
    bools = %w[ originally_ordered originally_trusted ordered trusted self_signed_untrusted ]
    attr_reader *attrs, *bools
    bools.each { |m| alias_method "#{m}?", m; private m }

    def to_s
      to_a.join
    end
    alias_method :to_a,   :resolved_chain
    alias_method :join,   :to_s
    alias_method :to_pem, :to_s
    alias_method :first,  :base_cert
  end
end
