# encoding: UTF-8
require 'openssl'
require_relative 'pem_scanner'
require_relative 'error'

module SSLTool
  module KeyHelper

    class KeyHelperError < ::SSLTool::Error; end

    class KeyMalformedError < KeyHelperError; end

    class KeyNotPresentError < KeyHelperError; end

    class KeyNotFoundError < KeyHelperError
      attr_reader :cert, :keys
      def initialize(cert, keys)
        @cert, @keys = cert, keys
        super("None of the given #{keys.length} keys match the certificate #{cert.subject}.")
      end
    end

    extend self

    def scan(s)
      PEMScanner.keys_from(s)
    end

    def key(s)
      error = -> { raise KeyMalformedError, "Key appears to be malformed, or is passphrase-protected." }
      case s
      when /\A-----BEGIN EC PRIVATE KEY/     then OpenSSL::PKey::EC.new(s, '').tap{|k| k.singleton_class.class_eval{ alias_method :private?, :private_key? }}
      when /\A-----BEGIN (RSA )?PRIVATE KEY/ then OpenSSL::PKey::RSA.new(s, '')
      else error[]
      end
    rescue OpenSSL::PKey::ECError, OpenSSL::PKey::RSAError => e
      error[]
    end

    def keys(*pems)
      pems.flatten.map { |s| key(s) }
    end

    def find_private_key_for_certificate(cert, keys)
      keys.find { |key| cert.check_private_key(key) }
    end

    def find_private_key_for_certificate!(cert, keys)
      raise KeyNotPresentError if keys.empty?
      key = find_private_key_for_certificate(cert, keys)
      return key if key
      raise KeyNotFoundError.new(cert, keys)
    end

  end
end
